% The MIT License (MIT)
% Copyright (c) 2020 Elmar Rueckert
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE. 

% **********************************************************
% * Exercise to the Probabilistic Machine Learning Lecture
% *
% *                       Probabilistic Trajectorie  
% *                       Dataset is from the Paper: https://ai-lab.science/wp/SciReps_HumanContacts.pdf
% *                       Rueckert, Elmar and Jernej Camernik and Jan Peters and Jan Babic (2016). Scientific Reports, Nature Publishing Group. .
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert@ai-lab.science
% *
% **********************************************************

clear all;
close all;
clc;

load('HumanTargetReachingDataset.mat');

%This dataset contains n = 16 trajectories or demonstrations. 
%of the left wrist and right wrist human target reaching motions. Only the
%y-coordinate (D=2) is given. See
%https://rob.ai-lab.science/wp/SciReps_HumanContacts_Supplement.pdf for the
%coordinate frame. 

%Variables in the dataset                   Dimesnion
%----------------------------------------------------
%leftwirstTrajectories_yCoord               n x Tmax=338 time steps at 100Hz
%rightwirstTrajectories_yCoord              n x Tmax=391
%targetIds                                  n x 1 \in {1,2,3}

n=16;
D=2;
%Note that the trajectories have different lengths! 

%% Exercise 1: 
%Compute the trajectory lengths of each trajectory and strore the result in the matrix Treal \in n x D
Treal = nan(n,D);

%% Exercise 2:
%Compute algined trajetories of length 
T = 100;
%Data={y_11, y_21, ..., y_Tn} \in Txn
DataLeftWrist = nan(T,n);
DataRightWrist = nan(T,n);

%% Exercise 3: 
%Visualize the data in two panels for the left and right wrist. Use
%lines(3) as line colors dependent on the targetId.

%% Exercise 4: 
%Compute radial basis functions of the form
%phi^[i]_t = exp(-1/(2h) (z_t - c_i)^2) where i = 1...M with 
M = 5;  
c = linspace(0,1,M);
h = 0.5*(c(2)-c(1))^2;
z = linspace(0,1,100);

% use the basis functions to compute a matrix A = [phi_1, ..., phi_T] \in TxM
A = nan(T,M);
% normalize A such that sum(A) = 1 per time step!

%plot the basis functions and test what happens when you change h!

%% Exercise 5:
% compute the feature weights w using the basis function matrix A. 
ws_LeftWrist = nan(M,n);
ws_RightWrist = nan(M,n);
%visulalize the weights and test what happens when you increas M!

%% Exercise 6: 
%Compute the probabilistic trajectory model. 
%p(w) = N(mu,S). 
mu_LeftWrist = nan(M,1);
S_LeftWrist = nan(M,M);
%visualize the prior distribution and explain how the indivdual weights are
%correlated in time!

%% Exercise 7:
%Use the learned prior model to predict the mean trajectory and variance. 
%p(tau) = integral p(tau|w)p(w) dw 
%Use the function shadedErrorBar.m to visdualize the model.  plot the 95%
%confidenz interval (2*sigma)
%What happens if you increase the data noise parameter: 
sigma_y = 0.01;


%% Exercise 9: Trajectory prediction / completion
%given are the 1-dim. observations of the right wrist y-coordinate in
%meter. Compute the Bayes Estimate using the Bayes Rule. 
y_obs       = [-0.08, -0.02, 0.1]';
t_obs       = [50, 60, 70]';
Sigma_obs   = 1e-3*eye(3);
%predict the right wrist trajectory using the learned model. visualize the
%observations and the predictions.
%what happens if only the first two data points are observed?
Ao = [];
for oId=1:length(t_obs)
    Ao = [Ao; A(t_obs(oId), :)];
end
y_predObs = Ao*mu_RightWrist;
[y_obs, y_predObs]

%parameter posterior p(w_obs) \propto p(obs | Ao w) p(w) := N(mu_w_obs, Sigma_w_obs) 
L           = S_RightWrist * Ao' * (Sigma_obs +  Ao * S_RightWrist * Ao')^(-1);
mu_w_obs    = mu_RightWrist + L * (y_obs - Ao*mu_RightWrist);
Sigma_w_obs = S_RightWrist - L * Ao * S_RightWrist;
%predict the trajctory given the observarions. Use the function shadedErrorBar.m 
%to visdualize the prior model, the dataset and the prediction given the
%observations. 

%% Bonus Point Exercise: Model the coupling between dimensions
%Define and compute a combined weight vector wC \in D M 
% wC = [w^[1]_1, ..., w^[1]_M, w^[2]_1, ..., w^[D]_M];
wC = nan(D * M, 1);
%Compute AC as combined basis function model
AC = nan(D * T, D * M);
% repeat exercises 6-8 and hand in a printed report. For good explainations
% and high quality plots 2 bonus points are given!
